from os import PathLike
from pathlib import Path
from typing import List

import geojson
from geojson import Point
import numpy
from shapely.geometry import LineString
import typepigeon

from packetraven.packets import APRSPacket
from packetraven.packets.tracks import APRSTrack, LocationPacketTrack, PredictedTrajectory

KML_STANDARD = '{http://www.opengis.net/kml/2.2}'


def write_packet_tracks(packet_tracks: List[LocationPacketTrack], filename: PathLike):
    """
    write packet tracks to file
    """

    if not isinstance(filename, Path):
        filename = Path(filename)
    if filename.suffix == '.txt':
        packets = []
        for packet_track_index, packet_track in enumerate(packet_tracks):
            packets.extend(packet_track)
        packets = sorted(packets)
        lines = [
            f'{packet.time:%Y-%m-%d %H:%M:%S %Z}: {packet.frame}'
            if isinstance(packet, APRSPacket)
            else f'{packet.time:%Y-%m-%d %H:%M:%S %Z}'
            for packet in packets
        ]
        with open(filename, 'w') as output_file:
            output_file.write('\n'.join(lines))
    elif filename.suffix in ['.geojson', '.json']:
        features = []
        for packet_track in packet_tracks:
            ascent_rates = numpy.round(packet_track.ascent_rates, 3)
            ground_speeds = numpy.round(packet_track.ground_speeds, 3)

            for packet_index, packet in enumerate(packet_track):
                properties = {
                    'time': f'{packet.time:%Y%m%d%H%M%S}',
                    'altitude': float(packet.coordinates[2]),
                    'ascent_rate': float(ascent_rates[packet_index]),
                    'ground_speed': float(ground_speeds[packet_index]),
                }

                properties.update(packet.attributes)

                features.append(
                    geojson.Feature(
                        geometry=geojson.Point(packet.coordinates.tolist()),
                        properties=properties,
                    )
                )

            properties = {
                'name': packet_track.name,
                'start_time': f'{packet_track.packets[0].time:%Y-%m-%d %H:%M:%S}',
                'end_time': f'{packet_track.packets[-1].time:%Y-%m-%d %H:%M:%S}',
                'duration': f'{packet_track.packets[-1].time - packet_track.packets[0].time}',
                **packet_track.attributes,
            }

            if isinstance(packet_track, PredictedTrajectory):
                properties.update(
                    {**packet_track.metadata, **packet_track.parameters,}
                )
            else:
                properties.update(
                    {
                        'seconds_to_ground': packet_track.time_to_ground,
                        'last_altitude': float(packet_track.coordinates[-1, -1]),
                        'last_ascent_rate': float(ascent_rates[-1]),
                        'last_ground_speed': float(ground_speeds[-1]),
                    }
                )

            features.append(
                geojson.Feature(
                    geometry=geojson.LineString(
                        [packet.coordinates.tolist() for packet in packet_track.packets]
                    ),
                    properties=typepigeon.convert_to_json(properties),
                )
            )

        features = geojson.FeatureCollection(features)

        with open(filename, 'w') as output_file:
            geojson.dump(features, output_file)
    elif filename.suffix == '.kml':
        from fastkml import kml

        output_kml = kml.KML()
        document = kml.Document(
            KML_STANDARD, '1', 'root document', 'root document, containing geometries'
        )
        output_kml.append(document)

        for packet_track_index, packet_track in enumerate(packet_tracks):
            ascent_rates = numpy.round(packet_track.ascent_rates, 3)
            ground_speeds = numpy.round(packet_track.ground_speeds, 3)

            for packet_index, packet in enumerate(packet_track):
                placemark = kml.Placemark(
                    KML_STANDARD,
                    f'1 {packet_track_index} {packet_index}',
                    f'{packet.time:%Y%m%d%H%M%S} {packet_track.callsign if isinstance(packet_track, APRSTrack) else ""}',
                    f'altitude={packet.coordinates[2]} '
                    f'ascent_rate={ascent_rates[packet_index]} '
                    f'ground_speed={ground_speeds[packet_index]}',
                )
                placemark.geometry = Point(packet.coordinates.tolist())
                document.append(placemark)

            properties = {
                'altitude': packet_track.coordinates[-1, -1],
                'duration': f'{packet_track.packets[-1].time - packet_track.packets[0].time}',
            }

            if isinstance(packet_track, PredictedTrajectory):
                properties.update(
                    {**packet_track.metadata, **packet_track.parameters,}
                )
            else:
                properties.update(
                    {
                        'seconds_to_ground': packet_track.time_to_ground,
                        'last_altitude': float(packet_track.coordinates[-1, -1]),
                        'last_ascent_rate': float(ascent_rates[-1]),
                        'last_ground_speed': float(ground_speeds[-1]),
                    }
                )

            properties = ' '.join(f'{key}={value}' for key, value in properties.items())

            placemark = kml.Placemark(
                KML_STANDARD, f'1 {packet_track_index}', packet_track.name, properties,
            )
            placemark.geometry = LineString(packet_track.coordinates)
            document.append(placemark)

        with open(filename, 'w') as output_file:
            output_file.write(output_kml.to_string())
    else:
        raise NotImplementedError(
            f'saving to file type "{filename.suffix}" has not been implemented'
        )



from abc import ABC, abstractmethod
from copy import deepcopy
from os import PathLike
from pathlib import Path
from typing import Any, Collection, Dict, Generator, Mapping

import typepigeon
import yaml

from packetraven.connections.base import PacketSource


class Configuration(ABC, Mapping):
    fields: Dict[str, type]
    defaults: Dict[str, Any] = None

    def __init__(self, **configuration):
        if not hasattr(self, 'configuration'):
            self.__configuration = {field: None for field in self.fields}
        if len(configuration) > 0:
            self.update(configuration)

        if self.defaults is not None:
            update_none(self.__configuration, self.defaults)

        missing_fields = [field for field in self.fields if field not in self.__configuration]
        if len(missing_fields) > 0:
            raise ValueError(
                f'missing {len(missing_fields)} fields required by "{self.__class__.__name__}" - {list(missing_fields)}'
            )

    @classmethod
    def from_file(cls, filename: PathLike) -> 'Configuration':
        raise NotImplementedError()

    def __copy__(self) -> 'Configuration':
        return self.__class__(**deepcopy(self.__configuration))

    def __contains__(self, key: str) -> bool:
        return key in self.__configuration

    def __getitem__(self, key: str) -> Any:
        return self.__configuration[key]

    def __setitem__(self, key: str, value: Any):
        if key in self.fields:
            field_type = self.fields[key]
            if not isinstance(field_type, type):
                field_type = type(field_type)
            if not isinstance(value, field_type):
                value = typepigeon.convert_value(value, self.fields[key])
        else:
            self.fields[key] = type(value)
        self.__configuration[key] = value

    def update(self, other: Mapping):
        for key, value in other.items():
            if key in self:
                if isinstance(self[key], Configuration):
                    self[key].update(value)
                else:
                    field_type = self.fields[key]
                    if (
                        isinstance(field_type, type)
                        and issubclass(field_type, Configuration)
                        and value is not None
                    ):
                        converted_value = field_type(**value)
                    elif isinstance(field_type, Mapping) and value is not None:
                        converted_value = convert_key_pairs(value, field_type)
                    else:
                        converted_value = typepigeon.convert_value(value, field_type)
                    if self[key] is None or self[key] != converted_value:
                        value = converted_value
            self[key] = value

    def __eq__(self, other: 'Configuration') -> bool:
        return other.__configuration == self.__configuration

    def __repr__(self):
        configuration = ', '.join(
            [f'{key}={repr(value)}' for key, value in self.__configuration.items()]
        )
        return f'{self.__class__.__name__}({configuration})'

    def __len__(self) -> int:
        return len(self.__configuration)

    def __iter__(self) -> Generator:
        yield from self.__configuration

    def __delitem__(self, key):
        del self.__configuration[key]

    @abstractmethod
    def to_file(self, filename: PathLike = None, overwrite: bool = False):
        raise NotImplementedError()


class ConfigurationSection:
    name: str


class ConfigurationYAML(Configuration):
    @classmethod
    def from_file(cls, filename: PathLike) -> 'Configuration':
        with open(filename) as input_file:
            configuration = yaml.safe_load(input_file)
        return cls(**configuration)

    def to_file(self, filename: PathLike = None, overwrite: bool = True):
        if not isinstance(filename, Path):
            filename = Path(filename)
        if overwrite or not filename.exists():
            content = typepigeon.convert_to_json(self._Configuration__configuration)
            with open(filename, 'w') as input_file:
                yaml.safe_dump(content, input_file)


class PacketSourceConfiguration(ABC):
    def packet_source(self, callsigns: [str] = None) -> PacketSource:
        raise NotImplementedError()


def convert_key_pairs(value_mapping: Mapping, type_mapping: Mapping[str, type]) -> Mapping:
    value_mapping = dict(**value_mapping)
    keys = list(value_mapping)
    for key in keys:
        if key in type_mapping:
            value = value_mapping[key]
            value_type = type_mapping[key]
            if not isinstance(value_type, (Collection)) and issubclass(
                value_type, Configuration
            ):
                value = value_type(**value)
            elif isinstance(value, dict):
                value = convert_key_pairs(value, value_type)
            else:
                value = typepigeon.convert_value(value, value_type)
        else:
            value = typepigeon.convert_value(value_mapping, type_mapping)
        value_mapping[key] = value
    return value_mapping


def update_none(values: Dict[str, Any], defaults: Dict[str, Any]):
    for key, default_value in defaults.items():
        if key not in values or values[key] is None:
            values[key] = default_value
        elif isinstance(values[key], Mapping):
            update_none(values[key], default_value)



from copy import copy
from datetime import datetime, timedelta
from os import PathLike
from typing import Any, Dict, Iterable, List, Union

from dateutil.parser import parse as parse_date
import geojson
import numpy
from pandas import DataFrame
from pyproj import CRS

from packetraven.model import (
    FREEFALL_DESCENT_RATE,
    FREEFALL_DESCENT_RATE_UNCERTAINTY,
    FREEFALL_SECONDS_TO_GROUND,
)
from packetraven.packets import APRSPacket, DEFAULT_CRS, LocationPacket
from packetraven.packets.base import Distance
from packetraven.packets.structures import DoublyLinkedList


class LocationPacketTrack:
    """
    a collection of location packets in chronological order
    """

    def __init__(
        self,
        packets: List[LocationPacket] = None,
        name: str = None,
        crs: CRS = None,
        **attributes,
    ):
        """
        :param packets: iterable of packets
        :param name: name of packet track
        :param crs: coordinate reference system to use
        """

        if name is None:
            name = 'packet track'
        elif not isinstance(name, str):
            name = str(name)

        self.name = name
        self.packets = DoublyLinkedList(None)
        self.crs = crs if crs is not None else DEFAULT_CRS
        self.attributes = attributes

        if packets is not None:
            self.extend(packets)

        self.__length = len(self)
        self.__data = None

    def append(self, packet: LocationPacket):
        if packet not in self.packets:
            if packet.crs != self.crs:
                packet.transform_to(self.crs)
            self.packets.append(packet)

    def extend(self, packets: List[LocationPacket]):
        for packet in packets:
            self.append(packet)

    def sort(self, inplace: bool = False):
        if inplace:
            instance = self
        else:
            instance = copy(self)
        instance.packets.sort(inplace=True)
        return instance

    @property
    def data(self) -> DataFrame:
        if self.__data is None or len(self) != self.__length:
            records = [
                {
                    'time': packet.time,
                    'x': packet.coordinates[0],
                    'y': packet.coordinates[1],
                    'altitude': packet.coordinates[2],
                }
                for packet in self.packets
            ]
            for index, packet_delta in enumerate(
                numpy.insert(numpy.diff(self.packets), 0, Distance(0, 0, 0, self.crs))
            ):
                records[index].update(
                    {
                        'interval': packet_delta.interval,
                        'overground_distance': packet_delta.overground,
                        'ascent': packet_delta.ascent,
                        'ground_speed': packet_delta.ground_speed,
                        'ascent_rate': packet_delta.ascent_rate,
                    }
                )

            self.__data = DataFrame.from_records(records)

        return self.__data

    @property
    def times(self) -> numpy.ndarray:
        return self.data['time'].values

    @property
    def coordinates(self) -> numpy.ndarray:
        return self.data[['x', 'y', 'altitude']].values

    @property
    def altitudes(self) -> numpy.ndarray:
        return self.coordinates[:, -1]

    @property
    def intervals(self) -> numpy.ndarray:
        return self.data['interval'].values

    @property
    def overground_distances(self) -> numpy.ndarray:
        """ overground distances between packets """
        return self.data['overground_distance'].values

    @property
    def ascents(self) -> numpy.ndarray:
        """ differences in altitude between packets """
        return self.data['ascent'].values

    @property
    def ascent_rates(self) -> numpy.ndarray:
        """ instantaneous ascent rates between packets """
        return self.data['ascent_rate'].values

    @property
    def ground_speeds(self) -> numpy.ndarray:
        """ instantaneous overground speeds between packets """
        return self.data['ground_speed'].values

    @property
    def cumulative_overground_distances(self) -> numpy.ndarray:
        """ cumulative overground distances from start """
        return numpy.cumsum(self.overground_distances)

    @property
    def time_to_ground(self) -> timedelta:
        """ estimated time to reach the ground at the current rate of descent """

        current_ascent_rate = self.ascent_rates[-1]
        if current_ascent_rate < 0:
            # TODO implement landing location as the intersection of the predicted descent track with a local DEM
            # TODO implement a time to impact calc based off of standard atmo
            return timedelta(seconds=self.altitudes[-1] / abs(current_ascent_rate))
        else:
            return timedelta(seconds=-1)

    @property
    def distance_downrange(self) -> float:
        """ direct overground distance between first and last packets only """
        if len(self.packets) > 0:
            return self.packets[-1].overground_distance(self.coordinates[0, :2])
        else:
            return 0.0

    @property
    def length(self) -> float:
        """ total length of the packet track over the ground """
        return sum([distance.overground for distance in self.packets.difference])

    def __getitem__(
        self, index: Union[int, Iterable[int], slice]
    ) -> Union[LocationPacket, 'LocationPacketTrack']:
        if isinstance(index, str):
            try:
                index = parse_date(index)
            except:
                index = int(index)
        if isinstance(index, int):
            return self.packets[index]
        elif isinstance(index, Iterable) or isinstance(index, slice):
            if isinstance(index, numpy.ndarray) and index.dtype == bool:
                index = numpy.where(index)[0]
            if isinstance(index, slice) or len(index) > 0:
                packets = self.packets[index]
            else:
                packets = None
            return self.__class__(packets=packets, crs=self.crs, **self.attributes)
        elif isinstance(index, (datetime, numpy.datetime64)):
            if not isinstance(index, numpy.datetime64):
                index = numpy.datetime64(index)
            matching_packets = []
            for packet in self.packets:
                if packet.time == index:
                    matching_packets.append(packet)
            if len(matching_packets) == 0:
                maximum_interval = max(self.intervals)
                if (
                    index > min(self.times) - maximum_interval
                    and index < max(self.times) + maximum_interval
                ):
                    smallest_difference = None
                    closest_time = None
                    for packet in self.packets:
                        packet_time = numpy.datetime64(packet.time)
                        difference = abs(packet_time - index)
                        if smallest_difference is None or difference < smallest_difference:
                            smallest_difference = difference
                            closest_time = packet_time
                    return self[closest_time]
                else:
                    raise IndexError(
                        f'time index out of range: {min(self.times) - maximum_interval} <\= {index} <\= {max(self.times) + maximum_interval}'
                    )
            elif len(matching_packets) == 1:
                return matching_packets[0]
            else:
                return self.__class__(packets=matching_packets, name=self.name, crs=self.crs)
        else:
            raise ValueError(f'unrecognized index: {index}')

    def __iter__(self):
        return iter(self.packets)

    def __reversed__(self):
        return reversed(self.packets)

    def __contains__(self, item) -> bool:
        return item in self.packets

    def __len__(self) -> int:
        return len(self.packets)

    def __eq__(self, other) -> bool:
        return self.packets == other.packets and self.attributes == other.attributes

    def __str__(self) -> str:
        return str(list(self))

    @property
    def dataframe(self) -> DataFrame:
        return DataFrame(
            {
                'name': [self.name for _ in range(len(self))],
                'times': self.times,
                'x': self.coordinates[:, 0],
                'y': self.coordinates[:, 1],
                'z': self.coordinates[:, 2],
                'intervals': self.intervals,
                'overground_distances': self.overground_distances,
                'ascents': self.ascents,
                'ascent_rates': self.ascent_rates,
                'ground_speeds': self.ground_speeds,
                'cumulative_overground_distances': self.cumulative_overground_distances,
            }
        )

    @classmethod
    def from_file(cls, filename: PathLike) -> List['LocationPacketTrack']:
        """
        load packet tracks from the given file

        :param filename: file path to GeoJSON
        :return: packet tracks
        """

        with open(filename) as input_file:
            data = geojson.load(input_file)

        points = []
        linestrings = []
        for feature in data['features']:
            if feature['geometry']['type'] == 'Point':
                points.append(feature)
            elif feature['geometry']['type'] == 'LineString':
                linestrings.append(feature)

        tracks = []
        for linestring in linestrings:
            packets = []
            for coordinate in linestring['geometry']['coordinates']:
                for index, point in enumerate(points):
                    if point['geometry']['coordinates'] == coordinate:
                        packet = LocationPacket(
                            x=point['geometry']['coordinates'][0],
                            y=point['geometry']['coordinates'][1],
                            z=point['geometry']['coordinates'][2],
                            **point['properties'],
                        )
                        packets.append(packet)
                        if index != 0:
                            points.pop(index)
                        break
            track = LocationPacketTrack(packets=packets, **linestring['properties'])

            tracks.append(track)

        return tracks


class BalloonTrack(LocationPacketTrack):
    """
    a packet track describing the path of a balloon
    """

    def __init__(
        self,
        packets: List[LocationPacket] = None,
        name: str = None,
        crs: CRS = None,
        **attributes,
    ):
        LocationPacketTrack.__init__(self, packets=packets, name=name, crs=crs, **attributes)
        self.__falling = False

    @property
    def time_to_ground(self) -> timedelta:
        current_ascent_rate = self.ascent_rates[-1]
        if current_ascent_rate < 0:
            if self.falling:
                return timedelta(seconds=FREEFALL_SECONDS_TO_GROUND(self.altitudes[-1]))
            else:
                current_altitude = self.altitudes[-1]
                # TODO implement landing location as the intersection of the predicted descent track with a local DEM
                return timedelta(seconds=current_altitude / -current_ascent_rate)
        else:
            return timedelta(seconds=-1)

    @property
    def falling(self) -> bool:
        current_ascent_rate = self.ascent_rates[-1]
        if current_ascent_rate >= 0:
            self.__falling = False
        elif not self.__falling:
            current_altitude = self.altitudes[-1]
            freefall_descent_rate = FREEFALL_DESCENT_RATE(current_altitude)
            freefall_descent_rate_uncertainty = FREEFALL_DESCENT_RATE_UNCERTAINTY(
                current_altitude
            )
            if numpy.abs(current_ascent_rate - freefall_descent_rate) < numpy.abs(
                freefall_descent_rate_uncertainty
            ):
                self.__falling = True
        return self.__falling


class APRSTrack(BalloonTrack):
    """
    a balloon track comprised of APRS packets
    """

    def __init__(
        self,
        packets: List[APRSPacket] = None,
        callsign: str = None,
        crs: CRS = None,
        **attributes,
    ):
        """
        :param packets: iterable of packets
        :param callsign: callsign of APRS packets
        :param crs: coordinate reference system to use
        """

        if 'name' not in attributes:
            attributes['name'] = 'APRS track'

        if callsign is not None:
            if not isinstance(callsign, str):
                callsign = str(callsign)
            if len(callsign) > 9 or ' ' in callsign:
                raise ValueError(f'unrecognized callsign format: "{callsign}"')

        BalloonTrack.__init__(self, packets=packets, callsign=callsign, crs=crs, **attributes)

    @property
    def callsign(self) -> str:
        return self.attributes['callsign']

    def append(self, packet: APRSPacket):
        packet_callsign = packet['callsign']

        if self.callsign is None or packet_callsign == self.callsign:
            super().append(packet)
        else:
            raise ValueError(
                f'Packet callsign {packet_callsign} does not match ground track callsign {self.callsign}.'
            )

    def __str__(self) -> str:
        return f'{self.callsign}: {super().__str__()}'

    @classmethod
    def from_file(cls, filename: PathLike) -> List['APRSTrack']:
        tracks = LocationPacketTrack.from_file(filename)

        for index in range(len(tracks)):
            track = tracks[index]
            packets = []
            for packet in track.packets:
                packets.append(
                    APRSPacket(
                        from_callsign=packet.attributes['from'],
                        to_callsign=packet.attributes['to'],
                        time=packet.time,
                        x=packet.coordinates[0],
                        y=packet.coordinates[1],
                        z=packet.coordinates[2],
                        **{
                            key: value
                            for key, value in packet.attributes.items()
                            if key not in ['from', 'to']
                        },
                    )
                )
            tracks[index] = APRSTrack(packets=packets, crs=track.crs, **track.attributes)

        return tracks


class PredictedTrajectory(LocationPacketTrack):
    """
    a prediction trajectory retrieved from a prediction API
    """

    def __init__(
        self,
        packets: List[LocationPacket],
        parameters: Dict[str, Any],
        metadata: Dict[str, Any],
        crs: CRS = None,
        **attributes,
    ):
        """
        :param packets: iterable of packets
        :param parameters: prediction parameters
        :param metadata: prediction completion metadata
        :param crs: coordinate reference system to use
        """

        if 'name' not in attributes:
            attributes['name'] = 'predicted trajectory'

        self.__parameters = parameters
        self.__metadata = metadata

        LocationPacketTrack.__init__(
            self, packets=packets, crs=crs, **attributes,
        )

    @property
    def landing_site(self) -> (float, float, float):
        return self.coordinates[-1]

    @classmethod
    def from_file(cls, filename: PathLike) -> List['PredictedTrajectory']:
        tracks = LocationPacketTrack.from_file(filename)

        for index in range(len(tracks)):
            track = tracks[index]
            tracks[index] = PredictedTrajectory(
                packets=track.packets, crs=track.crs, **track.attributes,
            )

        return tracks

    @property
    def parameters(self) -> Dict[str, Any]:
        return self.__parameters

    @property
    def metadata(self) -> Dict[str, Any]:
        return self.__metadata



from tkinter import Toplevel
from typing import Dict

import matplotlib

matplotlib.use('TkAgg')
matplotlib.interactive(True)

from matplotlib import pyplot
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from packetraven.packets.tracks import LocationPacketTrack, PredictedTrajectory

VARIABLES = {
    'altitude': {'x': 'times', 'y': 'altitudes', 'xlabel': 'time', 'ylabel': 'altitude (m)'},
    'ascent_rate': {
        'x': 'times',
        'y': 'ascent_rates',
        'xlabel': 'time',
        'ylabel': 'ascent rate (m/s)',
    },
    'ground_speed': {
        'x': 'ground_speeds',
        'y': 'altitudes',
        'xlabel': 'ground speed (m/s)',
        'ylabel': 'altitude (m)',
    },
}


class LiveTrackPlot:
    """
    wrapper for a `matplotlib` plot window that can be updated on the fly with packet tracks
    """

    def __init__(
        self,
        packet_tracks: Dict[str, LocationPacketTrack],
        variable: str,
        predictions: Dict[str, PredictedTrajectory] = None,
    ):
        if variable not in VARIABLES:
            raise NotImplementedError(f'unsupported plotting variable "{variable}"')

        self.packet_tracks = packet_tracks
        self.predictions = predictions if predictions is not None else {}
        self.variable = variable

        try:
            self.window.protocol('WM_DELETE_WINDOW', self.window.iconify)
        except AttributeError:
            pass

        self.update()

    def update(
        self,
        packet_tracks: Dict[str, LocationPacketTrack] = None,
        predictions: Dict[str, PredictedTrajectory] = None,
    ):
        if packet_tracks is not None:
            self.packet_tracks.update(packet_tracks)
        if predictions is not None:
            self.predictions.update(predictions)

        if len(self.packet_tracks) > 0 or len(self.predictions) > 0:
            try:
                if self.window.state() == 'iconic':
                    self.window.deiconify()
                if self.window.focus_get() is None:
                    self.window.focus_force()
            except AttributeError:
                pass

            axis = self.axis

            while len(axis.lines) > 0:
                axis.lines.pop(-1)

            packet_track_lines = {}
            for name, packet_track in self.packet_tracks.items():
                x = getattr(packet_track, VARIABLES[self.variable]['x'])
                y = getattr(packet_track, VARIABLES[self.variable]['y'])
                lines = axis.plot(x, y, linewidth=2, marker='o', label=packet_track.name,)

                packet_track_lines[name] = lines[0]

            for name, packet_track in self.predictions.items():
                color = (
                    packet_track_lines[name].get_color()
                    if name in packet_track_lines
                    else None
                )

                axis.plot(
                    getattr(packet_track, VARIABLES[self.variable]['x']),
                    getattr(packet_track, VARIABLES[self.variable]['y']),
                    '--',
                    linewidth=0.5,
                    color=color,
                    label=f'{packet_track.name} prediction',
                )

            axis.legend()

            # NOTE: this `pyplot.pause(0.1)` NEEDS to be here, and it NEEDS to be `0.1` seconds;
            # otherwise the plot does not render correctly upon update
            pyplot.pause(0.1)

    @property
    def window(self) -> Toplevel:
        return self.figure.canvas.manager.window

    @property
    def figure(self) -> Figure:
        if not hasattr(self, '__figure') or not pyplot.fignum_exists(self.__figure.number):
            self.__figure = self.__new_figure()
        return self.__figure

    @property
    def axis(self) -> Axes:
        if not hasattr(self, '__axis') or self.__axis.figure is not self.figure:
            self.__axis = self.__new_axis()
        return self.__axis

    def __new_figure(self) -> Figure:
        x_label = VARIABLES[self.variable]['xlabel']
        y_label = VARIABLES[self.variable]['ylabel']

        return pyplot.figure(num=f'{y_label} / {x_label}')

    def __new_axis(self, figure: Figure = None) -> Axes:
        if figure is None:
            figure = self.figure

        x_label = VARIABLES[self.variable]['xlabel']
        y_label = VARIABLES[self.variable]['ylabel']

        axis = figure.add_subplot(1, 1, 1)
        axis.set_xlabel(x_label)
        axis.set_ylabel(y_label)

        return axis

    def close(self):
        pyplot.close(self.figure.number)



from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Dict, Optional, Tuple, Union

import pytz
import requests
from shapely.geometry import Point
import typepigeon

from packetraven.packets import LocationPacket
from packetraven.packets.tracks import PredictedTrajectory

DEFAULT_FLOAT_ALTITUDE_UNCERTAINTY = 500
UTC_TIMEZONE = pytz.utc


class PredictionAPIURL(Enum):
    cusf = 'https://predict.cusf.co.uk/api/v1/'
    lukerenegar = 'https://predict.lukerenegar.com/api/v1.1/'


class FlightProfile(Enum):
    standard = 'standard_profile'
    float = 'float_profile'


class BalloonPredictionQuery(ABC):
    """
    balloon prediction API query
    """

    def __init__(
        self,
        api_url: str,
        start_location: Union[Tuple[float, float, Optional[float]], Point],
        start_time: datetime,
        ascent_rate: float,
        burst_altitude: float,
        sea_level_descent_rate: float,
        float_altitude: float = None,
        float_duration: timedelta = None,
        name: str = None,
        descent_only: bool = False,
    ):
        """
        :param api_url: URL of API
        :param start_location: location of balloon launch
        :param start_time: date and time of balloon launch
        :param ascent_rate: average ascent rate (m/s)
        :param burst_altitude: altitude at which balloon will burst
        :param sea_level_descent_rate: descent rate at sea level (m/s)
        :param float_altitude: altitude of float (m)
        :param float_duration: date and time of float end
        :param name: name of prediction track
        :param descent_only: whether to query for descent only
        """

        if not isinstance(start_location, Point):
            start_location = typepigeon.convert_value(start_location, Point)

        if name is None:
            name = 'prediction'

        if start_time is not None:
            if not isinstance(start_time, datetime):
                start_time = typepigeon.convert_value(start_time, datetime)
            if start_time.tzinfo is None or start_time.tzinfo.utcoffset(start_time) is None:
                start_time = UTC_TIMEZONE.localize(start_time)

        if float_duration is not None:
            if not isinstance(float_duration, datetime):
                float_duration = typepigeon.convert_value(float_duration, timedelta)

        self.api_url = api_url
        self.start_location = start_location
        self.start_time = start_time
        self.ascent_rate = ascent_rate
        self.burst_altitude = burst_altitude
        self.sea_level_descent_rate = abs(sea_level_descent_rate)
        self.float_altitude = float_altitude
        self.float_duration = float_duration
        self.name = name
        self.descent_only = descent_only

    @property
    @abstractmethod
    def query(self) -> Dict[str, Any]:
        raise NotImplementedError

    def get(self) -> Dict[str, Any]:
        response = requests.get(self.api_url, params=self.query)
        return response.json()

    @property
    @abstractmethod
    def predict(self) -> PredictedTrajectory:
        raise NotImplementedError

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({repr(self.api_url)}, {repr(self.start_location)}, {repr(self.start_time)}, {repr(self.ascent_rate)}, {repr(self.burst_altitude)}, {repr(self.sea_level_descent_rate)})'


class PredictionError(Exception):
    pass


class CUSFBalloonPredictionQuery(BalloonPredictionQuery):
    """
    connection to https://predict.cusf.co.uk/api/v1/

    >>> cusf_api = CUSFBalloonPredictionQuery(start_location=(-77.547824, 39.359031), launch_datetime=datetime.now(), ascent_rate=5.5, burst_altitude=28000, descent_rate=9)
    >>> predicted_track = cusf_api.predict
    """

    def __init__(
        self,
        start_location: Union[Tuple[float, float, Optional[float]], Point],
        start_time: datetime,
        ascent_rate: float,
        burst_altitude: float,
        sea_level_descent_rate: float,
        profile: FlightProfile = None,
        version: float = None,
        dataset_time: datetime = None,
        float_altitude: float = None,
        float_duration: timedelta = None,
        api_url: PredictionAPIURL = None,
        name: str = None,
        descent_only: bool = False,
    ):
        if profile is None:
            if not descent_only and (float_altitude is not None or float_duration is not None):
                profile = FlightProfile.float
            else:
                profile = FlightProfile.standard

        if dataset_time is not None:
            if not isinstance(dataset_time, datetime):
                dataset_time = typepigeon.convert_value(dataset_time, datetime)
            if (
                dataset_time.tzinfo is None
                or dataset_time.tzinfo.utcoffset(dataset_time) is None
            ):
                dataset_time = UTC_TIMEZONE.localize(dataset_time)

        if api_url is None:
            api_url = PredictionAPIURL.cusf

        if isinstance(api_url, PredictionAPIURL):
            api_url = api_url.value

        if name is None:
            name = 'cusf_prediction'

        super().__init__(
            api_url,
            start_location,
            start_time,
            ascent_rate,
            burst_altitude,
            sea_level_descent_rate,
            float_altitude,
            float_duration,
            name,
            descent_only,
        )

        # CUSF API requires longitude in 0-360 format
        if self.start_location.x < 0:
            x = self.start_location.x + 360
        else:
            x = self.start_location.x
        launch_coordinates = [x, self.start_location.y]
        if self.start_location.has_z:
            launch_coordinates.append(self.start_location.z)
        self.launch_site = Point(launch_coordinates)

        self.profile = profile if not self.descent_only else FlightProfile.standard
        self.version = version
        self.dataset_time = dataset_time

    @property
    def query(self) -> Dict[str, Any]:
        query = {
            'launch_longitude': self.launch_site.x,
            'launch_latitude': self.launch_site.y,
            'launch_datetime': self.start_time.isoformat(),
            'ascent_rate': self.ascent_rate,
            'burst_altitude': self.burst_altitude
            if not self.descent_only
            else self.launch_site.z + 1,
            'descent_rate': self.sea_level_descent_rate,
        }

        if self.launch_site.has_z:
            query['launch_altitude'] = self.launch_site.z
        if self.profile is not None:
            query['profile'] = self.profile.value
        if self.version is not None:
            query['version'] = self.version
        if self.dataset_time is not None:
            query['dataset'] = self.dataset_time.isoformat()

        if self.profile == FlightProfile.float and not self.descent_only:
            if self.float_altitude is None:
                self.float_altitude = self.burst_altitude
            if self.float_duration is None:
                raise PredictionError('float duration not provided')
            query['float_altitude'] = self.float_altitude
            start_altitude = self.start_location.z if self.start_location.has_z else 0
            float_start_time = self.start_time + timedelta(
                seconds=(self.float_altitude - start_altitude) / self.ascent_rate
            )
            query['stop_datetime'] = (float_start_time + self.float_duration).isoformat()

        return query

    def get(self) -> Dict[str, Any]:
        response = requests.get(self.api_url, params=self.query)

        if response.status_code == 200:
            response = response.json()
            if 'error' not in response:
                # TODO tawhiri currently does not include descent when querying a float profile
                if self.profile == FlightProfile.float:
                    # this code runs another prediction query with a standard profile and extracts the descent stage to append to the response from the original query
                    for stage in response['prediction']:
                        # if a descent stage exists, we don't need to do anything
                        if stage['stage'] == 'descent':
                            break
                    else:
                        for stage in response['prediction']:
                            if stage['stage'] == 'float':
                                float_end = stage['trajectory'][-1]
                                break
                        else:
                            raise PredictionError('API did not return a float trajectory')

                        standard_profile_query = self.__class__(
                            start_location=[
                                float_end['longitude'],
                                float_end['latitude'],
                                float_end['altitude'],
                            ],
                            start_time=typepigeon.convert_value(
                                float_end['datetime'], datetime
                            ),
                            ascent_rate=10,
                            burst_altitude=float_end['altitude'] + 0.1,
                            sea_level_descent_rate=self.sea_level_descent_rate,
                            profile=FlightProfile.standard,
                            version=self.version,
                            dataset_time=self.dataset_time,
                            float_altitude=None,
                            float_duration=None,
                            api_url=self.api_url,
                            name=self.name,
                            descent_only=True,
                        )

                        for stage in standard_profile_query.get()['prediction']:
                            if stage['stage'] == 'descent':
                                response['prediction'].append(stage)
                                break

                if self.descent_only:
                    indices_to_remove = []
                    for index, stage in enumerate(response['prediction']):
                        # if a descent stage exists, we don't need to do anything
                        if stage['stage'] != 'descent':
                            indices_to_remove.append(index)
                            break
                    for index in indices_to_remove:
                        response['prediction'].pop(index)

                return response
            else:
                raise PredictionError(response['error']['description'])
        else:
            try:
                error = response.json()['error']['description']
            except:
                error = 'no message'
            raise ConnectionError(
                f'connection raised error {response.status_code} for {response.url} - {error}'
            )

    @property
    def predict(self) -> PredictedTrajectory:
        response = self.get()

        if 'error' not in response:
            points = []

            for stage in response['prediction']:
                points.extend(stage['trajectory'])

            for point in points:
                if point['longitude'] > 180:
                    point['longitude'] -= 360

            return PredictedTrajectory(
                packets=[
                    LocationPacket(
                        point['datetime'],
                        point['longitude'],
                        point['latitude'],
                        point['altitude'],
                    )
                    for point in points
                ],
                parameters=response['request'],
                metadata=response['metadata'],
            )
        else:
            raise PredictionError(response['error']['description'])


class LukeRenegarBalloonPredictionQuery(CUSFBalloonPredictionQuery):
    """
    connection to https://predict.lukerenegar.com/api/v1.1/
    """

    def __init__(
        self,
        start_location: Union[Tuple[float, float, Optional[float]], Point],
        start_time: datetime,
        ascent_rate: float,
        burst_altitude: float,
        sea_level_descent_rate: float,
        ascent_rate_standard_deviation: float = None,
        burst_altitude_standard_deviation: float = None,
        descent_rate_standard_deviation: float = None,
        wind_standard_deviation: float = None,
        use_monte_carlo: bool = None,
        physics_model: str = None,
        profile: FlightProfile = None,
        version: float = None,
        dataset_time: datetime = None,
        float_altitude: float = None,
        float_duration: timedelta = None,
        api_url: str = None,
        name: str = None,
        descent_only: bool = False,
    ):
        if api_url is None:
            api_url = PredictionAPIURL.lukerenegar

        if name is None:
            name = 'lrenegar_prediction'

        super().__init__(
            start_location,
            start_time,
            ascent_rate,
            burst_altitude,
            sea_level_descent_rate,
            profile,
            version,
            dataset_time,
            float_altitude,
            float_duration,
            api_url,
            name,
            descent_only,
        )

        self.ascent_rate_standard_deviation = ascent_rate_standard_deviation
        self.burst_altitude_standard_deviation = burst_altitude_standard_deviation
        self.descent_rate_standard_deviation = descent_rate_standard_deviation
        self.wind_standard_deviation = wind_standard_deviation
        self.use_monte_carlo = use_monte_carlo
        self.physics_model = physics_model

    @property
    def query(self) -> Dict[str, Any]:
        query = super().query

        if self.ascent_rate_standard_deviation is not None:
            query['ascent_rate_std_dev'] = self.ascent_rate_standard_deviation
        if self.burst_altitude_standard_deviation is not None:
            query['burst_altitude_std_dev'] = self.burst_altitude_standard_deviation
        if self.descent_rate_standard_deviation is not None:
            query['descent_rate_std_dev'] = self.descent_rate_standard_deviation
        if self.wind_standard_deviation is not None:
            query['wind_std_dev'] = self.wind_standard_deviation
        if self.use_monte_carlo is not None:
            query['monte_carlo'] = self.use_monte_carlo
        if self.physics_model is not None:
            query['physics_model'] = self.physics_model

        return query



import logging
import os

from tablecrow.utilities import parse_hostname

from packetraven.configuration.base import (
    ConfigurationSection,
    ConfigurationYAML,
    PacketSourceConfiguration,
)
from packetraven.connections.internet import APRSfi, APRSis, PacketDatabaseTable


class SSHCredentials(ConfigurationYAML, ConfigurationSection):
    name = 'tunnel'
    fields = {
        'hostname': str,
        'port': int,
        'username': str,
        'password': str,
    }
    defaults = {
        'port': 22,
    }

    @classmethod
    def from_connection_string(cls, connection_string: str):
        configuration = {}
        if connection_string is not None:
            configuration = parse_hostname(connection_string)
        return cls(**configuration)


class DatabaseCredentials(ConfigurationYAML, ConfigurationSection, PacketSourceConfiguration):
    name = 'database'
    fields = {
        'hostname': str,
        'port': int,
        'database': str,
        'table': str,
        'username': str,
        'password': str,
        'tunnel': SSHCredentials,
    }
    defaults = {
        'hostname': os.getenv('POSTGRES_HOSTNAME', 'localhost'),
        'port': os.getenv('POSTGRES_PORT', 5432),
        'database': os.getenv('POSTGRES_DATABASE', 'postgres'),
        'username': os.getenv('POSTGRES_USERNAME', 'postgres'),
        'password': os.getenv('POSTGRES_PASSWORD', ''),
    }

    @classmethod
    def from_connection_string(cls, connection_string: str, tunnel: SSHCredentials = None):
        configuration = {}
        if tunnel is not None:
            configuration['tunnel'] = tunnel
        if connection_string is not None:
            parsed = parse_hostname(connection_string)

            hostname = parsed['hostname']
            port = parsed['port']
            username = parsed['username']
            password = parsed['password']

            if hostname.count('/') != 2:
                logging.warning(f'unable to parse connection information from "{parsed}"')

            (
                configuration['hostname'],
                configuration['database'],
                configuration['table'],
            ) = hostname.split('/')

            configuration['username'] = username
            configuration['password'] = password
            configuration['port'] = port

        return cls(**configuration)

    def packet_source(self, callsigns: [str] = None) -> PacketDatabaseTable:
        ssh_kwargs = {}
        if self['tunnel'] is not None:
            ssh_kwargs.update({f'ssh_{key}': value for key, value in self['tunnel'].items()})

        return PacketDatabaseTable(
            hostname=self['hostname'],
            database=self['database'],
            table=self['table'],
            callsigns=callsigns,
            **ssh_kwargs,
        )


class APRSfiCredentials(ConfigurationYAML, ConfigurationSection, PacketSourceConfiguration):
    name = 'aprs_fi'
    fields = {
        'api_key': str,
    }
    defaults = {
        'api_key': os.getenv('APRS_FI_API_KEY', None),
    }

    def packet_source(self, callsigns: [str] = None) -> APRSfi:
        return APRSfi(callsigns=callsigns, api_key=self['api_key'])


class APRSisCredentials(ConfigurationYAML, ConfigurationSection, PacketSourceConfiguration):
    name = 'aprs_is'
    fields = {
        'hostname': str,
    }

    def packet_source(self, callsigns: [str] = None) -> APRSis:
        return APRSis(callsigns=callsigns, hostname=self['hostname'])


class CredentialsYAML(ConfigurationYAML):
    fields = {
        'aprs_fi': APRSfiCredentials,
        'database': DatabaseCredentials,
        'aprs_is': APRSisCredentials,
    }

    def __init__(self, **configuration):
        for section in [APRSfiCredentials, DatabaseCredentials, APRSisCredentials]:
            if section.name in configuration:
                configuration[section.name] = section(**configuration[section.name])
        super().__init__(**configuration)



from copy import copy
from datetime import datetime, timedelta
from getpass import getpass
import logging
from pathlib import Path
import sys
import time
from typing import Dict, List

from dateutil.tz import tzlocal
import humanize as humanize
import numpy
import typer

from packetraven.configuration.prediction import PredictionConfiguration
from packetraven.configuration.run import RunConfiguration
from packetraven.connections import (
    PacketDatabaseTable,
    PacketGeoJSON,
    RawAPRSTextFile,
    SerialTNC,
    TimeIntervalError,
)
from packetraven.connections.base import PacketSource
from packetraven.packets import APRSPacket
from packetraven.packets.tracks import APRSTrack, LocationPacketTrack, PredictedTrajectory
from packetraven.packets.writer import write_packet_tracks
from packetraven.predicts import CUSFBalloonPredictionQuery, PredictionError

logging.basicConfig(
    stream=sys.stdout, level=logging.INFO, format='%(asctime)s | %(levelname)-8s | %(message)s'
)

DEFAULT_INTERVAL_SECONDS = 20


def packetraven_command(configuration_filename: str, gui: bool = False):
    """
    start parsing packets

    :param configuration_filename: configuration file in YAML format - see `examples` directory for examples
    :param gui: start the graphical interface
    """

    if configuration_filename is not None:
        configuration = RunConfiguration.from_file(configuration_filename)

        if configuration['log'] is not None:
            logging.basicConfig(filename=configuration['log']['filename'])

        filter_message = 'retrieving packets'
        if configuration['time']['start'] is not None and configuration['time']['end'] is None:
            filter_message += f' sent after {configuration["time"]["start"]:%Y-%m-%d %H:%M:%S}'
        elif (
            configuration['time']['start'] is None and configuration['time']['end'] is not None
        ):
            filter_message += f' sent before {configuration["time"]["end"]:%Y-%m-%d %H:%M:%S}'
        elif (
            configuration['time']['start'] is not None
            and configuration['time']['end'] is not None
        ):
            filter_message += (
                f' sent between {configuration["time"]["start"]:%Y-%m-%d %H:%M:%S}'
                f' and {configuration["time"]["end"]:%Y-%m-%d %H:%M:%S}'
            )
        if configuration['callsigns'] is not None:
            filter_message += f' from {len(configuration["callsigns"])} callsigns: {configuration["callsigns"]}'
        logging.info(filter_message)

        if configuration['callsigns'] is not None:
            aprsfi_url = (
                f'https://aprs.fi/#!call=a%2F{"%2Ca%2F".join(configuration["callsigns"])}'
            )
            if configuration['time']['start'] is not None:
                aprsfi_url += f'&ts={configuration["time"]["start"]:%s}'
            if configuration['time']['end'] is not None:
                aprsfi_url += f'&te={configuration["time"]["end"]:%s}'
            logging.info(f'tracking URL: {aprsfi_url}')
    else:
        configuration = None

    if gui:
        from packetraven.gui import PacketRavenGUI

        PacketRavenGUI(configuration)
    elif configuration is not None:
        connections = []
        if 'text' in configuration['packets']:
            for location in configuration['packets']['text']['locations']:
                location = location.strip()
                try:
                    if Path(location).suffix in ['.txt', '.log']:
                        connection = RawAPRSTextFile(location, configuration['callsigns'])
                        logging.info(f'reading text file {connection.location}')
                    elif Path(location).suffix in ['.json', '.geojson']:
                        connection = PacketGeoJSON(
                            location, callsigns=configuration['callsigns']
                        )
                        logging.info(f'reading GeoJSON file {connection.location}')
                    else:
                        connection = SerialTNC(location, configuration['callsigns'])
                        logging.info(f'opened port {connection.location} for APRS parsing')
                    connections.append(connection)
                except ConnectionError as error:
                    logging.warning(f'{error.__class__.__name__} - {error}')

        if 'aprs_fi' in configuration['packets']:
            try:
                aprs_api = configuration['packets']['aprs_fi'].packet_source(
                    callsigns=configuration['callsigns']
                )
                logging.info(f'connected to {aprs_api.location}')
                connections.append(aprs_api)
            except ConnectionError as error:
                logging.warning(f'{error.__class__.__name__} - {error}')

        if 'database' in configuration['packets']:
            database_credentials = configuration['packets']['database']
            ssh_tunnel_credentials = database_credentials['tunnel']
            try:
                if ssh_tunnel_credentials is not None:
                    if ssh_tunnel_credentials['username'] is None:
                        ssh_username = input(
                            f'enter username for SSH host "{ssh_tunnel_credentials["hostname"]}": '
                        )
                        if ssh_username is None or len(ssh_username) == 0:
                            raise ConnectionError('missing SSH username')
                        ssh_tunnel_credentials['username'] = ssh_username

                    if ssh_tunnel_credentials['password'] is None:
                        ssh_password = getpass(
                            f'enter password for SSH user "{ssh_tunnel_credentials["username"]}": '
                        )
                        if ssh_password is None or len(ssh_password) == 0:
                            raise ConnectionError('missing SSH password')
                        ssh_tunnel_credentials['password'] = ssh_password

                if database_credentials['username'] is None:
                    database_username = input(
                        f'enter username for database '
                        f'"{database_credentials["hostname"]}/{database_credentials["database"]}": '
                    )
                    if database_username is None or len(database_username) == 0:
                        raise ConnectionError('missing database username')
                    database_credentials['username'] = database_username

                if database_credentials['password'] is None:
                    database_password = getpass(
                        f'enter password for database user "{database_credentials["username"]}": '
                    )
                    if database_password is None or len(database_password) == 0:
                        raise ConnectionError('missing database password')
                    database_credentials['password'] = database_password

                database = database_credentials.packet_source(
                    callsigns=configuration['callsigns']
                )
                logging.info(f'connected to {database.location}')
                connections.append(database)
            except ConnectionError:
                database = None
        else:
            database = None

        if len(connections) == 0:
            if (
                configuration['output']['filename'] is not None
                and configuration['output']['filename'].exists()
            ):
                connections.append(PacketGeoJSON(configuration['output']['filename']))
            else:
                logging.error(f'no connections started')
                sys.exit(1)

        if 'aprs_is' in configuration['packets']:
            try:
                aprs_is = configuration['packets']['aprs_is'].packet_source(
                    callsigns=configuration['callsigns']
                )
            except ConnectionError:
                aprs_is = None
        else:
            aprs_is = None

        logging.info(
            f'listening for packets every {configuration["time"]["interval"]}s from {len(connections)} connection(s): '
            f'{", ".join([connection.location for connection in connections])}'
        )

        packet_tracks = {}
        predictions = {}
        plots = {}

        if any(configuration['plots'].values()):
            from packetraven.gui.plotting import LiveTrackPlot

            for variable, enabled in configuration['plots'].items():
                if enabled and variable not in plots:
                    plots[variable] = LiveTrackPlot(packet_tracks, variable, predictions)
                elif not enabled and variable in plots:
                    plots[variable].close()
                    del plots[variable]

        try:
            time_without_packets = timedelta(seconds=0)
            while len(connections) > 0:
                current_time = datetime.now()

                try:
                    new_packets = retrieve_packets(
                        connections,
                        packet_tracks,
                        database,
                        start_date=configuration['time']['start'],
                        end_date=configuration['time']['end'],
                    )

                    output_filename_index = None
                    for index, connection in enumerate(connections):
                        if isinstance(connection, PacketGeoJSON):
                            output_filename_index = index
                    if output_filename_index is not None:
                        connections.pop(output_filename_index)

                    if configuration['prediction'] is not None:
                        try:
                            predictions.update(
                                packet_track_predictions(
                                    packet_tracks, configuration['prediction'],
                                )
                            )
                            write_packet_tracks(
                                predictions.values(),
                                configuration['prediction']['output']['filename'],
                            )
                        except PredictionError as error:
                            logging.warning(f'{error.__class__.__name__} - {error}')
                        except Exception as error:
                            logging.warning(
                                f'error retrieving prediction trajectory - {error.__class__.__name__} - {error}'
                            )
                except Exception as error:
                    logging.exception(f'{error.__class__.__name__} - {error}')
                    new_packets = {}

                if len(new_packets) > 0:
                    if configuration['output']['filename'] is not None:
                        write_packet_tracks(
                            [packet_tracks[callsign] for callsign in packet_tracks],
                            configuration['output']['filename'],
                        )
                    if aprs_is is not None:
                        for packets in new_packets.values():
                            aprs_is.send(packets)

                    for plot in plots.values():
                        plot.update(packet_tracks, predictions)
                else:
                    time_without_packets += configuration['time']['interval'] + (
                        datetime.now() - current_time
                    )

                if (
                    'timeout' in configuration['time']
                    and time_without_packets >= configuration['time']['timeout']
                ):
                    logging.info(
                        f'shutting down - no packets received for {time_without_packets}'
                    )
                    break
                else:
                    time.sleep(configuration['time']['interval'] / timedelta(seconds=1))
        except KeyboardInterrupt:
            for connection in connections:
                connection.close()
            sys.exit(0)
    else:
        logging.error(
            'no configuration given; to use the graphical interface try `packetraven --gui`'
        )
        sys.exit(1)


def retrieve_packets(
    connections: List[PacketSource],
    packet_tracks: List[LocationPacketTrack],
    database: PacketDatabaseTable = None,
    start_date: datetime = None,
    end_date: datetime = None,
) -> Dict[str, APRSPacket]:
    logging.debug(f'receiving packets from {len(connections)} source(s)')
    current_time = datetime.now()

    parsed_packets = []
    for connection in connections:
        try:
            connection_packets = connection.packets
            parsed_packets.extend(connection_packets)
        except ConnectionError as error:
            logging.error(f'{connection.__class__.__name__} - {error}')
        except TimeIntervalError:
            pass

    logging.debug(f'received {len(parsed_packets)} packets')

    new_packets = {}
    if len(parsed_packets) > 0:
        updated_callsigns = set()
        for parsed_packet in parsed_packets:
            callsign = parsed_packet['callsign']

            start_date = ensure_datetime_timezone(start_date)
            end_date = ensure_datetime_timezone(end_date)
            parsed_packet.time = ensure_datetime_timezone(parsed_packet.time)

            if start_date is not None and parsed_packet.time < start_date:
                continue
            if end_date is not None and parsed_packet.time > end_date:
                continue

            if callsign not in packet_tracks:
                packet_tracks[callsign] = APRSTrack(packets=[parsed_packet], callsign=callsign)
                logging.debug(f'started tracking callsign {callsign:8}')
            else:
                packet_track = packet_tracks[callsign]
                if parsed_packet not in packet_track:
                    packet_track.append(parsed_packet)
                else:
                    if database is None or parsed_packet.source != database.location:
                        logging.debug(f'skipping duplicate packet: {parsed_packet}')
                    continue

            if parsed_packet.source not in new_packets:
                new_packets[parsed_packet.source] = []
            new_packets[parsed_packet.source].append(parsed_packet)
            if callsign not in updated_callsigns:
                updated_callsigns.add(callsign)

        for source in new_packets:
            new_packets[source] = list(sorted(new_packets[source]))

        for source, packets in new_packets.items():
            logging.info(f'received {len(packets)} new packet(s) from {source}')

        if database is not None:
            for packets in new_packets.values():
                database.send(
                    packet for packet in packets if packet.source != database.location
                )

        updated_callsigns = sorted(updated_callsigns)
        for callsign in updated_callsigns:
            packet_track = packet_tracks[callsign]
            packet_time = datetime.utcfromtimestamp(
                (packet_track.times[-1] - numpy.datetime64('1970-01-01T00:00:00'))
                / numpy.timedelta64(1, 's')
            )
            packet_track.sort(inplace=True)
            try:
                coordinate_string = ', '.join(
                    f'{coordinate:.3f}°' for coordinate in packet_track.coordinates[-1, :2]
                )
                logging.info(
                    f'{callsign:9} - packet #{len(packet_track):<3} - ({coordinate_string}, {packet_track.coordinates[-1, 2]:9.2f}m)'
                    f'; packet time is {packet_time} ({humanize.naturaltime(current_time - packet_time)}, {packet_track.intervals[-1] / numpy.timedelta64(1, "s"):6.1f} s interval)'
                    f'; traveled {packet_track.overground_distances[-1]:6.1f} m ({packet_track.ground_speeds[-1]:5.1f} m/s) over the ground'
                    f', and {packet_track.ascents[-1]:6.1f} m ({packet_track.ascent_rates[-1]:5.1f} m/s) vertically, since the previous packet'
                )
            except Exception as error:
                logging.exception(f'{error.__class__.__name__} - {error}')

        for callsign in updated_callsigns:
            packet_track = packet_tracks[callsign]
            packet_time = datetime.utcfromtimestamp(
                (packet_track.times[-1] - numpy.datetime64('1970-01-01T00:00:00'))
                / numpy.timedelta64(1, 's')
            )
            try:
                ascent_rates = packet_track.ascent_rates[packet_track.ascent_rates >= 0]
                if len(ascent_rates) > 0:
                    mean_ascent_rate = numpy.nanmean(ascent_rates)
                else:
                    mean_ascent_rate = 0

                descent_rates = packet_track.ascent_rates[packet_track.ascent_rates < 0]
                if len(descent_rates) > 0:
                    mean_descent_rate = numpy.nanmean(descent_rates)
                else:
                    mean_descent_rate = 0

                if len(packet_track.ground_speeds) > 0:
                    mean_ground_speed = numpy.nanmean(packet_track.ground_speeds)
                else:
                    mean_ground_speed = 0

                if len(packet_track.intervals) > 0:
                    mean_interval = numpy.nanmean(
                        packet_track.intervals / numpy.timedelta64(1, 's')
                    )
                else:
                    mean_interval = 0

                message = (
                    f'{callsign:9} - '
                    f'altitude: {packet_track.altitudes[-1]:6.1f} m'
                    f'; avg. ascent rate: {mean_ascent_rate:5.1f} m/s'
                    f'; avg. descent rate: {mean_descent_rate:5.1f} m/s'
                    f'; avg. ground speed: {mean_ground_speed:5.1f} m/s'
                    f'; avg. packet interval: {mean_interval:6.1f} s'
                )

                if packet_track.time_to_ground >= timedelta(seconds=0):
                    landing_time = packet_time + packet_track.time_to_ground
                    time_to_ground = current_time - landing_time
                    message += (
                        f'; estimated landing: {landing_time:%Y-%m-%d %H:%M:%S} ({humanize.naturaltime(time_to_ground)})'
                        f'; max altitude: {packet_track.coordinates[:, 2].max():.2f} m'
                    )

                logging.info(message)

            except Exception as error:
                logging.exception(f'{error.__class__.__name__} - {error}')

    return new_packets


def packet_track_predictions(
    packet_tracks: Dict[str, LocationPacketTrack], configuration=PredictionConfiguration,
) -> Dict[str, PredictedTrajectory]:
    """
    retrieve location tracks detailing predicted trajectory of balloon flight(s) from the current track location

    :param packet_tracks: location packet tracks
    :param configuration: prediction configuration
    """

    prediction_tracks = {}
    for name, packet_track in packet_tracks.items():
        track_configuration = copy(configuration)

        ascent_rates = packet_track.ascent_rates[packet_track.ascent_rates > 0]
        if len(ascent_rates) > 0:
            average_ascent_rate = numpy.mean(ascent_rates)
            if average_ascent_rate > 0:
                track_configuration['profile']['ascent_rate'] = average_ascent_rate

            # if packet track is descending, override given burst altitude
            if len(ascent_rates) > 2 and all(ascent_rates[-2:] < 0):
                track_configuration['profile']['burst_altitude'] = (
                    packet_track.altitudes[-1] + 1
                )

        last_packet = packet_track[-1]

        track_configuration['start']['time'] = last_packet.time

        track_configuration['start']['location'] = last_packet.coordinates
        if len(track_configuration['start']['location'].shape) > 1:
            track_configuration['start']['location'] = track_configuration['start'][
                'location'
            ][0, :]

        if track_configuration['float']['altitude'] is not None and not packet_track.falling:
            packets_at_float_altitude = packet_track[
                numpy.abs(track_configuration['float']['altitude'] - packet_track.altitudes)
                < track_configuration['float']['uncertainty']
            ]
            if (
                len(packets_at_float_altitude) > 0
                and packets_at_float_altitude[-1].time == packet_track.times[-1]
            ):
                track_configuration['float']['start_time'] = packets_at_float_altitude[0].time
                track_configuration['profile']['descent_only'] = False
            elif packet_track.ascent_rates[-1] >= 0:
                track_configuration['float']['start_time'] = track_configuration['start'][
                    'time'
                ] + timedelta(
                    seconds=(
                        track_configuration['float']['altitude']
                        - track_configuration['start']['location'][2]
                    )
                    / track_configuration['profile']['ascent_rate']
                )
                track_configuration['profile']['descent_only'] = False
            else:
                track_configuration['float']['start_time'] = None
                track_configuration['profile']['descent_only'] = True
                track_configuration['profile'][
                    'sea_level_descent_rate'
                ] = packet_track.ascent_rates[-1]
        else:
            track_configuration['profile']['descent_only'] = packet_track.falling or numpy.any(
                packet_track.ascent_rates[-2:] < 0
            )

        try:
            prediction_query = CUSFBalloonPredictionQuery(
                start_location=track_configuration['start']['location'],
                start_time=track_configuration['start']['time'],
                ascent_rate=track_configuration['profile']['ascent_rate'],
                burst_altitude=track_configuration['profile']['burst_altitude'],
                sea_level_descent_rate=track_configuration['profile'][
                    'sea_level_descent_rate'
                ],
                float_altitude=track_configuration['float']['altitude'],
                float_duration=track_configuration['float']['duration'],
                api_url=track_configuration['api_url'],
                name=name,
                descent_only=track_configuration['profile']['descent_only'],
            )

            prediction = prediction_query.predict

            if packet_track.time_to_ground >= timedelta(seconds=0):
                logging.info(
                    f'{packet_track.name:<9} - predicted landing location: {prediction.coordinates[-1]} - https://www.google.com/maps/place/{prediction.coordinates[-1][1]},{prediction.coordinates[-1][0]}'
                )

            prediction_tracks[name] = prediction
        except Exception as error:
            logging.warning(
                f'error retrieving prediction trajectory - {error.__class__.__name__} - {error}'
            )

    if all(
        value is not None
        for value in [
            configuration['start']['location'],
            configuration['start']['time'],
            configuration['profile']['ascent_rate'],
            configuration['profile']['burst_altitude'],
            configuration['profile']['sea_level_descent_rate'],
        ]
    ):
        name = 'flight'

        try:
            prediction_query = CUSFBalloonPredictionQuery(
                start_location=configuration['start']['location'],
                start_time=configuration['start']['time'],
                ascent_rate=configuration['profile']['ascent_rate'],
                burst_altitude=configuration['profile']['burst_altitude'],
                sea_level_descent_rate=configuration['profile']['sea_level_descent_rate'],
                float_altitude=configuration['float']['altitude'],
                float_duration=configuration['float']['duration'],
                api_url=configuration['api_url'],
                name=name,
                descent_only=configuration['profile']['descent_only'],
            )

            prediction = prediction_query.predict
            prediction_tracks[name] = prediction
        except Exception as error:
            logging.warning(
                f'error retrieving prediction trajectory - {error.__class__.__name__} - {error}'
            )

    return prediction_tracks


def ensure_datetime_timezone(value: datetime) -> datetime:
    if value is not None and (value.tzinfo is None or value.tzinfo.utcoffset(value) is None):
        value = value.replace(tzinfo=tzlocal())
    return value


def main():
    typer.run(packetraven_command)


if __name__ == '__main__':
    main()



from datetime import datetime, timedelta
from pathlib import Path
from typing import Any

from packetraven.configuration.base import ConfigurationYAML
from packetraven.configuration.credentials import (
    APRSfiCredentials,
    APRSisCredentials,
    DatabaseCredentials,
)
from packetraven.configuration.prediction import PredictionConfiguration
from packetraven.configuration.text import TextStreamConfiguration

DEFAULT_INTERVAL_SECONDS = 20


class RunConfiguration(ConfigurationYAML):
    fields = {
        'callsigns': [str],
        'time': {
            'start': datetime,
            'end': datetime,
            'interval': timedelta,
            'timeout': timedelta,
        },
        'output': {'filename': Path,},
        'log': {'filename': Path,},
        'packets': {
            'aprs_fi': APRSfiCredentials,
            'text': TextStreamConfiguration,
            'database': DatabaseCredentials,
            'aprs_is': APRSisCredentials,
        },
        'prediction': PredictionConfiguration,
        'plots': {'altitude': bool, 'ascent_rate': bool, 'ground_speed': bool,},
    }

    defaults = {
        'callsigns': [],
        'time': {
            'start': None,
            'end': None,
            'interval': timedelta(seconds=DEFAULT_INTERVAL_SECONDS),
        },
        'output': {'filename': None},
        'log': {'filename': None},
        'packets': {},
        'plots': {'altitude': False, 'ascent_rate': False, 'ground_speed': False,},
    }

    def __setitem__(self, key: str, value: Any):
        super().__setitem__(key, value)

        if key == 'callsigns':
            for index, entry in enumerate(self['callsigns']):
                if ',' in entry:
                    del self['callsigns'][index]
                    for callsign in reversed(entry.split(',')):
                        self['callsigns'].insert(index, callsign)

        if key == 'time':
            if (
                'start' in self['time']
                and self['time']['start'] is not None
                and 'start' in self['time']
                and self['time']['end'] is not None
            ):
                if self['time']['start'] > self['time']['end']:
                    temp_start_date = self['time']['start']
                    self['time']['start'] = self['time']['end']
                    self['time']['end'] = temp_start_date
                    del temp_start_date

            if (
                'interval' in self['time']
                and self['time']['interval'] is not None
                and self['time']['interval'] < timedelta(seconds=1)
            ):
                self['time']['interval'] = timedelta(seconds=1)

        if key == 'output':
            output_filename = self['output']['filename']
            if output_filename is not None:
                output_filename = Path(output_filename).expanduser()
                if output_filename.is_dir() or (
                    not output_filename.exists() and output_filename.suffix == ''
                ):
                    output_filename = (
                        output_filename
                        / f'packetraven_output_{datetime.now():%Y%m%dT%H%M%S}.geojson'
                    )
                if not output_filename.parent.exists():
                    output_filename.parent.mkdir(parents=True, exist_ok=True)
                self['output']['filename'] = output_filename

        if key == 'log':
            log_filename = self['log']['filename']
            if log_filename is not None:
                log_filename = Path(log_filename).expanduser()
                if log_filename.is_dir() or (
                    not log_filename.exists() and log_filename.suffix == ''
                ):
                    log_filename = (
                        log_filename / f'packetraven_log_{datetime.now():%Y%m%dT%H%M%S}.txt'
                    )
                if not log_filename.parent.exists():
                    log_filename.parent.mkdir(parents=True, exist_ok=True)
                self['log']['filename'] = log_filename



from datetime import datetime
import logging
from os import PathLike
from pathlib import Path
from typing import List
from urllib.parse import urlparse

import geojson
import requests
import typepigeon

from packetraven.connections.base import APRSPacketSource, PacketSource, TimeIntervalError
from packetraven.packets import APRSPacket, LocationPacket


class RawAPRSTextFile(APRSPacketSource):
    """
    connection to a text file where each line consists of the time sent (`YYYY-MM-DDTHH:MM:SS`) followed by a colon `:` and then a raw APRS string
    """

    def __init__(self, filename: PathLike = None, callsigns: str = None):
        """
        :param filename: path to text file
        :param callsigns: list of callsigns to return from source
        """

        if not urlparse(str(filename)).scheme in ['http', 'https', 'ftp', 'sftp']:
            if not isinstance(filename, Path):
                if isinstance(filename, str):
                    filename = filename.strip('"')
                filename = Path(filename)
            filename = str(filename)

        super().__init__(filename, callsigns=callsigns)
        self.__last_access_time = None
        self.__parsed_lines = []

    @property
    def packets(self) -> List[APRSPacket]:
        if self.__last_access_time is not None and self.interval is not None:
            interval = datetime.now() - self.__last_access_time
            if interval < self.interval:
                raise TimeIntervalError(
                    f'interval {interval} less than minimum interval {self.interval}'
                )

        if Path(self.location).exists():
            file_connection = open(Path(self.location).expanduser().resolve())
            lines = file_connection.readlines()
        else:
            file_connection = requests.get(self.location, stream=True)
            lines = file_connection.iter_lines()

        packets = []
        for line in lines:
            if len(line) > 0:
                if isinstance(line, bytes):
                    line = line.decode()
                if line not in self.__parsed_lines:
                    self.__parsed_lines.append(line)
                    try:
                        packet_time, raw_aprs = line.split(': ', 1)
                        packet_time = typepigeon.convert_value(packet_time, datetime)
                    except:
                        raw_aprs = line
                        packet_time = datetime.now()
                    raw_aprs = raw_aprs.strip()
                    try:
                        packets.append(
                            APRSPacket.from_frame(raw_aprs, packet_time, source=self.location)
                        )
                    except Exception as error:
                        logging.debug(f'{error.__class__.__name__} - {error}')

        file_connection.close()

        if self.callsigns is not None:
            packets = [packet for packet in packets if packet.from_callsign in self.callsigns]
        self.__last_access_time = datetime.now()

        return packets

    def close(self):
        pass

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self.location)}, {repr(self.callsigns)})'


class PacketGeoJSON(PacketSource):
    """
    connection to a GeoJSON file containing packet locations as points
    """

    def __init__(self, filename: PathLike = None, callsigns: [str] = None):
        """
        :param filename: path to GeoJSON file
        :param callsigns: list of callsigns to return from source
        """

        if not urlparse(str(filename)).scheme in ['http', 'https', 'ftp', 'sftp']:
            if not isinstance(filename, Path):
                if isinstance(filename, str):
                    filename = filename.strip('"')
                filename = Path(filename)
            filename = str(filename)

        super().__init__(filename, callsigns)
        self.__last_access_time = None

    @property
    def packets(self) -> List[LocationPacket]:
        if self.__last_access_time is not None and self.interval is not None:
            interval = datetime.now() - self.__last_access_time
            if interval < self.interval:
                raise TimeIntervalError(
                    f'interval {interval} less than minimum interval {self.interval}'
                )

        if Path(self.location).exists():
            with open(Path(self.location).expanduser().resolve()) as file_connection:
                features = geojson.load(file_connection)
        else:
            response = requests.get(self.location, stream=True)
            features = geojson.loads(response.text)

        packets = []
        for feature in features['features']:
            if feature['geometry']['type'] == 'Point':
                properties = feature['properties']
                time = typepigeon.convert_value(properties['time'], datetime)
                del properties['time']

                if 'from' in properties:
                    from_callsign = properties['from']
                    to_callsign = properties['to']
                    del properties['from'], properties['to']

                    packet = APRSPacket(
                        from_callsign,
                        to_callsign,
                        time,
                        *feature['geometry']['coordinates'],
                        source=self.location,
                        **properties,
                    )
                else:
                    packet = LocationPacket(
                        time,
                        *feature['geometry']['coordinates'],
                        source=self.location,
                        **properties,
                    )

                packets.append(packet)

        self.__last_access_time = datetime.now()

        return packets

    def close(self):
        pass

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self.location)})'



from datetime import datetime, timedelta
from os import PathLike
from pathlib import Path
from typing import Any, Dict, List, Mapping

import typepigeon
import yaml

from packetraven.configuration.base import ConfigurationSection, ConfigurationYAML
from packetraven.model import FREEFALL_DESCENT_RATE
from packetraven.packets.tracks import PredictedTrajectory
from packetraven.predicts import (
    BalloonPredictionQuery,
    DEFAULT_FLOAT_ALTITUDE_UNCERTAINTY,
    PredictionAPIURL,
)


class PredictionConfiguration(ConfigurationYAML, ConfigurationSection):
    name = 'prediction'
    fields = {
        'start': {'location': [float], 'time': datetime},
        'profile': {
            'ascent_rate': float,
            'burst_altitude': float,
            'sea_level_descent_rate': float,
            'descent_only': bool,
        },
        'float': {'altitude': float, 'uncertainty': float, 'duration': timedelta},
        'output': {'filename': Path},
        'api_url': str,
        'name': str,
    }
    defaults = {
        'profile': {
            'sea_level_descent_rate': FREEFALL_DESCENT_RATE(0),
            'descent_only': False,
        },
        'float': {
            'altitude': None,
            'uncertainty': DEFAULT_FLOAT_ALTITUDE_UNCERTAINTY,
            'duration': None,
        },
        'output': {'filename': None},
        'api_url': PredictionAPIURL.cusf.value,
        'name': 'prediction',
    }

    def __init__(self, **configuration):
        super().__init__(**configuration)
        if self['float']['duration'] is not None:
            if self['float']['altitude'] is None:
                self['float']['altitude'] = self['profile']['burst_altitude']
            if self['float']['duration'] is None:
                raise ValueError('`float_duration` was not provided')

    def __setitem__(self, key: str, value: Any):
        super().__setitem__(key, value)

        if key == 'output':
            if self['output'] is not None and self['output']['filename'] is not None:
                output_filename = Path(self['output']['filename']).expanduser()
                if (
                    output_filename.is_dir()
                    or not output_filename.exists()
                    and output_filename.suffix == ''
                ):
                    output_filename /= (
                        f'packetraven_predict_{datetime.now():%Y%m%dT%H%M%S}.geojson'
                    )
                if not output_filename.parent.exists():
                    output_filename.parent.mkdir(parents=True, exist_ok=True)
                self['output']['filename'] = output_filename
        if key == 'start':
            if self['start'] is not None:
                if len(self['start']['location']) == 2:
                    self['start']['location'] = (*self['start']['location'], 0)
        if key in ['start', 'profile']:
            if self['start'] is not None and self['profile'] is not None:
                if self['profile']['burst_altitude'] <= self['start']['location'][2]:
                    self['profile']['burst_altitude'] = self['start']['location'][2] + 1

    @property
    def prediction(self) -> PredictedTrajectory:
        query = BalloonPredictionQuery(
            api_url=self['api_url'],
            start_location=self['start']['location'],
            start_time=self['start']['time'],
            ascent_rate=self['profile']['ascent_rate'],
            burst_altitude=self['profile']['burst_altitude'],
            sea_level_descent_rate=self['profile']['sea_level_descent_rate'],
            float_altitude=self['float']['altitude'],
            float_duration=self['float']['duration'],
            name=self['name'],
            descent_only=self['profile']['descent_only'],
        )
        return query.predict


class PredictionCloudConfiguration(ConfigurationYAML):
    fields = {
        'default': PredictionConfiguration,
        'perturbations': Dict[str, PredictionConfiguration],
    }

    def __init__(
        self,
        default: PredictionConfiguration,
        perturbations: List[PredictionConfiguration],
        **configuration,
    ):
        if not isinstance(perturbations, Mapping):
            perturbations = {
                f'profile_{index + 1}': perturbation
                for index, perturbation in enumerate(
                    typepigeon.convert_value(perturbations, [PredictionConfiguration])
                )
            }

        configuration['default'] = default
        configuration['perturbations'] = perturbations

        super().__init__(**configuration)

    @classmethod
    def from_file(cls, filename: PathLike) -> 'PredictionCloudConfiguration':
        with open(filename) as input_file:
            configuration = yaml.safe_load(input_file)

        default = configuration['default']
        for perturbation, perturbation_values in configuration['perturbations'].items():
            for default_key, default_value in default.items():
                if default_key not in perturbation_values:
                    configuration['perturbations'][perturbation][default_key] = default_value

        return cls(**configuration)

    def to_file(self, filename: PathLike = None, overwrite: bool = True):
        if not isinstance(filename, Path):
            filename = Path(filename)
        if overwrite or not filename.exists():
            configuration = self._Configuration__configuration

            default = configuration['default']
            for perturbation in configuration['perturbations'].values():
                for key in list(perturbation):
                    if key in default and perturbation[key] == default[key]:
                        del perturbation[key]

            content = typepigeon.convert_to_json(configuration)
            with open(filename, 'w') as input_file:
                yaml.safe_dump(content, input_file)



from pathlib import Path
from typing import List, Union
import warnings

from packetraven import RawAPRSTextFile, SerialTNC
from packetraven.configuration.base import (
    ConfigurationSection,
    ConfigurationYAML,
    PacketSourceConfiguration,
)
from packetraven.connections import PacketGeoJSON


class TextStreamConfiguration(ConfigurationYAML, ConfigurationSection):
    name = 'text'
    fields = {
        'locations': [str],
    }
